/*-
 * AvpMonitor for Linux (v0.1) 
 * Thanks Markus Wolf <klog@hert.org>, Promisc Security, who gives
 * us basic idea of interceptor.
 *
 * Copyright (C) 1999 Kaspersky Lab.
 * http://www.avp.ru
 *
 * This file is part of AvpMonitor
 * 
 * AvpMonitor is free software; you can redistribute it and/or modify
 * it without limitation.
 *
 * AvpMonitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GNU CC; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.  
 */

#include <stdio.h> 
#include <string.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <ctype.h>

#include "common.h"
#include "../audit.h" 
#include "profile.h"
#include "report.h"

void die(char *msg)
 {
//  printf("Error: %s",msg);
  perror(msg);
  exit(-1);
 }

#ifdef DEBUG_LOG      // only if enabled 
// *===========================================================*
// *     ==>    the logging utility
// * This procedure is used to log info (mostly debugging ones)
// * the log file is quite large, actually, so be careful....
// *-----------------------------------------------------------
int Log(int LogLevel,char *fmt,...)
 {
  if(LogLevel>DEBUG_LOG_LEVEL) return 0;
  va_list args;

  va_start(args,fmt);     // but the second argument is not used... 
  vfprintf(stdout,fmt,args);
  return 1;
 }
#endif

char* CheckSymlink(char *path,dev_t my_dev);

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>        // struct sockaddr_un 
#include <sys/file.h>
#include <time.h>
#include <stdarg.h>
#include <paths.h>

#define AVP_NODE_DIR      "/var/run"
#define AVP_NODE_PID      AVP_NODE_DIR "/AvpPid"
#define AVP_NODE_CTL      AVP_NODE_DIR "/AvpCtl\0\0\0"
char *NodePid=AVP_NODE_PID;
char *NodeCtl=AVP_NODE_CTL;
int DaemonPid=-1;

static int CheckPresent(void)
 {
  struct sockaddr_un CtlAddr;
  int CtlFd,pid,len,ret=-1;
  
  bzero((char *)&CtlAddr,sizeof(CtlAddr));
  CtlAddr.sun_family=AF_UNIX;
  strcpy(CtlAddr.sun_path,NodeCtl);
  if((CtlFd=socket(AF_UNIX,SOCK_STREAM,0))<0)
   {
    LOG((0,"Checking present. socket(%s) not created\n",CtlAddr.sun_path));
    return -1;
   }    
  len=sizeof(CtlAddr.sun_family)+strlen(NodeCtl);
  if(connect(CtlFd,(struct sockaddr *)(&CtlAddr),len)>=0)
   {
    // AvpDaemon is runnin, get its pid 
    FILE *f=fopen(AVP_NODE_PID,"r");
    if(f && fscanf(f,"%i",&pid)==1)
     {
      LOG((1,"Checking present: AvpLinuxDaemon is running as pid %i\n",pid));
      ret=pid;
     }
    else
     {
      LOG((0,"Checking present. Fatal error (Socket server exists, bat %s not open or not read)\n",AVP_NODE_PID));
      ret=-2;//exit(1);
     }
   }
  else 
   LOG((0,"Checking present. Fatal error (Socket server not exists)\n"));

  // not connected, so I'm alone 
  close(CtlFd);
  return ret;
 }

int TestObjectByDaemon(char log,int hlog,char *ObjName)
 {
  static int  AvpSrv=-1;		// fd for log 
  struct sockaddr_un AvpTestAddr;	// AF_UNIX address of local logger 
  int rez=-1;

  reportPutS(log,hlog,ObjName);

  //if(AvpSrv==-1)
   {
    bzero((char *)&AvpTestAddr,sizeof(AvpTestAddr));
    AvpTestAddr.sun_family=AF_UNIX;
    strcpy(AvpTestAddr.sun_path,NodeCtl);
    if((AvpSrv=socket(AF_UNIX,SOCK_STREAM,0))<0)
     {
      LOG((0,"create socket error: socket() not created\n"));
      return rez;
     } 
   }

  if(AvpSrv!=-1 &&
    connect(AvpSrv,(struct sockaddr *)(&AvpTestAddr),sizeof(AvpTestAddr.sun_family)+strlen(NodeCtl))>=0)
   {
    register int cnt;
    char tbuf[2048];
    time_t now;

    LOG((1,"AvpTest connected to AvpDaemon\n")); 
    
    // build the message 
    (void)time(&now);
    (void)sprintf(tbuf,"<%d>%.15s:",0,ctime(&now)+4);
    (void)strcpy(&tbuf[strlen(tbuf)],ObjName);

    cnt=strlen(tbuf);

    // output the message to the local logger 
    LOG((1,"write string (%s) to AvpDaemon\n",tbuf));
    if(write(AvpSrv,tbuf,cnt+1)>=0)
     {
      int Rez; 
      if((Rez=read(AvpSrv,tbuf,0x200))==-1)
       {
        LOG((0,"Cannot reaf from AvpDaemon!\n"));
        rez=1;
       }
      else 
       {
        tbuf[Rez]=0;
        LOG((0,"Test rezult: %s :%d\n",ObjName,atoi(tbuf) ));
        switch (atoi(tbuf))
	 {
	  case 10: break;
	  case  7: reportPutS(log,hlog,"\t 7 - File AvpLinux is corrupted"

); break;
	  case  5: reportPutS(log,hlog,"\t 5 - All viruses deleted
"); break;
	  case  4: reportPutS(log,hlog,"\t 4 - Known viruses were detected
"); break;
	  case  3: reportPutS(log,hlog,"\t 3 - Suspicious objects were found
"); break;
	  case  1: reportPutS(log,hlog,"\t 1 - Virus scan was not complete
"); break;
	  case  0: reportPutS(log,hlog,"\t 0 - No viruses were found"); break;
	  default: reportPutS(log,hlog,"\t Error!"); break;
	 }
        rez=2;
       }
     }
   }    
  else 
    LOG((0,"AvpTest cannot connected to AvpDaemon\n")); 

  (void)close(AvpSrv);
  LOG((2,"AvpTest close\n")); 
  reportPutS(log,hlog," \n");
  return(rez);
 }

int TestObject(char log,int hlog,char *szObjName)
 {
  struct stat st;
  if(lstat(szObjName,&st)==-1)
   {
    LOG((0,"Cannot fihd object (%s). LStat()=-1\n",szObjName));
    return -2;
   }    
  LOG((2,"Try to check type of test object %s \n",szObjName));
  if(S_ISLNK(st.st_mode))
   {
    if((prf_data.Symlinks==1)||(prf_data.Symlinks==2))
     {                                        
      char *TargetPath=CheckSymlink(szObjName,st.st_dev);//dev);
      LOG((1,"Type: symlink %s -> %s\n",szObjName,TargetPath));
      if(TargetPath!=NULL)
       if(lstat(TargetPath,&st)!=-1)
        if(S_ISDIR(st.st_mode))
	 {
          LOG((1,"Type: directory. Try to test directory %s from symlink %s \n",TargetPath,szObjName));
//        if(prf_data.Directories==1) ScanTree(TargetPath,GetFlags(),st.st_dev);
	  return 0;
	 }
        else
         {
          if(S_ISREG(st.st_mode) )//&& ((!fStats.st_size && verbose>4)|| fStats.st_size)
	   {
  	    LOG((1,"Type: regular file. Try to test file %s from symlink %s \n",TargetPath,szObjName));
            return TestObjectByDaemon(log,hlog,szObjName);
	   }    
 	 }
     } 
   }  
  else
    if(S_ISDIR(st.st_mode))
     {
      LOG((1,"Type: directory. Try to test %s \n",szObjName));
//    if(prf_data.Directories==1) ScanTree(szObjName,GetFlags(),st.st_dev); 
      return 5;
     }
    else
     {
      if(S_ISREG(st.st_mode) )//&& ((!fStats.st_size && verbose>4)|| fStats.st_size)
       {
        LOG((1,"Type: regular file. Try to test %s \n",szObjName));
        return TestObjectByDaemon(log,hlog,szObjName);
       }	
     }
  return -1;
 }

#define  U(c) ((c) & 0377) 
#define  QUOT    '\\'       

int SyInSy(unsigned smin, unsigned smy, unsigned smax)
 {
  char left[2],right[2],middle [2];

  left  [0]=smin; left  [1]='\0';
  right [0]=smax; right [1]='\0';
  middle[0]=smy;  middle[1]='\0';

  return (strcoll(left,middle)<=0 && strcoll(middle,right)<=0);
 }

int Match(register char *teststr,register char *mask)
 {
  register int scc; // current char 
  int c,cc,lc;   //lc previos sumbol in list
  int ok,notflag;

  for(;;)
   {
    scc=U(*teststr++);          
    switch (c=U(*mask++))
     { 
      case QUOT:          
        c=U(*mask++);
        if(c==0) return(0); 
        else goto def;
      case '[':   
        ok=notflag = 0;
        lc=077777;      
        if(*mask== '!')
         { notflag=1; mask++; }
        while((cc=U(*mask++))
!=0)
         {
          if(cc==']')
           {    
            if(ok)  break;      
            return (0);     
           }
          if(cc=='-')
           {    
            if(notflag)
             {
              if(!SyInSy (lc, scc,U(*mask++)))
 ok++;
              else return (0);
             }
            else
             {
              if(SyInSy(lc,scc,U(*mask++)))
 ok++;
             }
           }
          else
           {
            if(cc == QUOT)
             {      /* [\[\]] */
              cc = U(*mask++);
              if(!cc) return(0);
             }
            if(notflag)
             {
              if(scc && scc!=(lc=cc))
 ok++;
              else return (0);
             }
            else
             {
              if(scc==(lc=cc)) ok++;
             }
           }
         }
        if(cc == 0)
         {    
          //printf("None ]\n")

;
          return (0);        
         }
        continue;
      case '*':   
        if(!*mask)  return (1);
        for(teststr--;*teststr;teststr++)
          if(Match(teststr,mask))
 return (1);
        return (0);
      case 0:
        return (scc == 0);
      default: def:
        if(c!=scc) return (0);
        continue;
      case '?':   
        if(scc==0) return (0);
        continue;
     }
   }
 }

#define BOOL unsigned long
#define FALSE 0
#define TRUE  1
BOOL cmpFileMask(char* sTestFile)
 {
  if(prf_data.ExcludeFiles==1)
   {
    char s[85],*span;
    strcpy(s,prf_data.ExcludeMask);
    span=strtok((char*)s,";:, ");
    while(span)
     {
      if(Match(sTestFile,span)
==1)
        return TRUE;
      span=strtok(NULL,";:, ");
     }
   }
  return FALSE;
 }


int getline(int fd, char *buf, int len)
 {
  char chr;
  int i;

  bzero(buf, len);
  for(i=0;i<len;i++)
   {
    if(read(fd, &chr, 1) == 0) return -1;
    if((chr=='\n')||(chr==0)) return i;
    buf[i]=chr;
   }
  buf[i]=0;
  return i; 
 }

// To be used to get both strings and integers 
int logpos;
int getunit(char *src, char *dst, int len)
 {
  int i;

  bzero(dst, len);
  for(i=0;i<(len-1);logpos++,i++)
   {
    if(src[logpos]==' ') { logpos++; break; }
    if(src[logpos]==0) { logpos++; break; }
    dst[i] = src[logpos];
   }
  return (atol(dst));
 }

struct Message
 {
  int flag;
  char pname[PNAME_LEN];
  int uid;	// There may be special chars (like '*') 
  int pid;
  char buf[256];
 };

int getStructByLine(char *buf,struct Message *Mess)
 {
  char tmp[20];

  logpos=0;

  Mess->flag=getunit(buf, tmp, sizeof(tmp));
  getunit(buf,Mess->pname, sizeof(Mess->pname));
  Mess->pid =getunit(buf, tmp, sizeof(tmp));
  Mess->uid =getunit(buf, tmp, sizeof(tmp));
  //getunit(buf,Mess->filename, sizeof(Mess->filename)); 
  strcpy(Mess->buf,&buf[logpos]);
  return 0;
 }

void logline(struct Message Mess)
 {
  if((Mess.flag < 'A')||(Mess.flag > 'z')) return; // crapped

  char szInfo[100];
  szInfo[0]=0;
  if(prf_data.UseTime)
   {
    char *date;
    time_t t;

    t = time(NULL);
    date = ctime(&t);
    date[strlen(date)-5] = 0; // remove the year 
    sprintf(szInfo,"%s", date);
   }
  sprintf(&szInfo[strlen(szInfo)],"Program %s(PID %d): ",Mess.pname,Mess.uid);

  switch(Mess.flag)
   {
    case AUDIT_EXEC_BEFORE:
      log_file(szInfo,!prf_data.CommonLog,prf_data.AtRunLog.log.logfd,Mess.buf, "exec (before)"); break;
    case AUDIT_EXEC_AFTER:
      log_file(szInfo,!prf_data.CommonLog,prf_data.AtRunLog.log.logfd,Mess.buf, "exec (after)"); break;
    case AUDIT_OPEN_BEFORE:
      log_file(szInfo,!prf_data.CommonLog,prf_data.FSOpenLog.log.logfd,Mess.buf, "open (before)"); break;
    case AUDIT_OPEN_AFTER:
      if((strcmp(Mess.pname,"AvpDaemon")!=0)&&(strcmp(Mess.pname,"AvpLinux")!=0))
       {
        //reportPutS(prf_data.CommonLog,prf_data.CommonLogfd,Mess.buf);
	if(prf_data.CommonLog)TestObject(prf_data.CommonLog,prf_data.CommonLogfd,Mess.buf);
	else                  TestObject(!prf_data.CommonLog,prf_data.FSOpenLog.log.logfd,Mess.buf);
       }	
      log_file(szInfo,!prf_data.CommonLog,prf_data.FSOpenLog.log.logfd,Mess.buf, "open (after)"); break;
    case AUDIT_MODINIT:
      log_file(szInfo,!prf_data.CommonLog,prf_data.ModInstLog.log.logfd,Mess.buf, "modinit"); break;
    case AUDIT_CONNECT:
      log_socket(szInfo,!prf_data.CommonLog,prf_data.NetOutLog.log.logfd,Mess.buf, "connect", "-->"); break;
    case AUDIT_ACCEPT:
      log_socket(szInfo,!prf_data.CommonLog,prf_data.NetInLog.log.logfd,Mess.buf, "accept", "<--"); break;
    //case AUDIT_LISTEN:
    //  log_socket(szInfo,!prf_data.CommonLog,.logfd.logfd,Mess.buf, "listen", NULL); break;
    case AUDIT_SETUID:
      log_setuid(szInfo,!prf_data.CommonLog,prf_data.SetUIdLog.log.logfd,Mess.buf, "setuid"); break;
   }
 }

unsigned char f_help=0;                   // 1 - out help screen

char *start_path;   
char prg_path[0x200];

void About(void)
 {
  printf("\n+-----------------------------------------------------+\n"
           "| AntiViral Toolkit Pro by Eugene Kaspersky for Linux |\n"
           "|                     (Monitor)                       |\n"
           "|           Copyright(C) Kaspersky Lab. 1998          |\n"
           "|                Version 3.0  beta 1                  |\n"
           "|                                                     |\n"
           "|                   Evaluation copy                   |\n"  
           "+-----------------------------------------------------+\n");
 }

void usage(void)
 {
  printf("\n"
"Usage: AvpMonitor [options] \n"
"Valid options are:\n" //(not all of them are implemented)
"    -LP no symlinks are followed        -C[-]  don't cross filesystem boundaries\n"
"    -LH symlinks on the command line    -W[T][A][-][+][=filename]  save report\n"

"        are followed                    
  -WT or -WT+ truncate existing file\n"
"    -LL all symlinks are followed         -WA or -WA+ appends to existing file\n"
"    -E[-] do not put time in the logs     -W+ or -WA+ extended report\n"
"    -XF=masks exclude files\n"
"    -XD=masks exclude directories\n"
//"    -O[-]  write OK messages            -*  check all files\n"
//"    -R[-]  do not scan subdirectories   \n"
"For all options '-' inverts the default meaning.\n"
"Press Enter to finish...\n");
  getchar();
  while(!feof(stdin) && (fgetc(stdin)!='\n')) ;
 }

int main(int argc, char *argv[])
 {
  //int conffd;
  int srcfd;
  char buf[1024];
  int count;
  unsigned int i,len;
  char CmdOpts[0x50];

  About();

  CmdOpts[0]=0;
  for(unsigned char z=1;z<argc;z++)
   {
    if(*argv[z]=='-')
      if((strlen(CmdOpts)+strlen(argv[z])+1)<0x100) strcat(CmdOpts,argv[z]);
   }

  if((start_path=getcwd(NULL,0))==NULL)
   {
    printf("getcwd() failed\n");
    exit(0);
   }

  if(!*start_path || start_path[strlen(start_path)-1]!='/') strcat(start_path,"/");

  strcpy(prg_path,*argv); // program name
  for(i=(unsigned short)(strlen(prg_path)-1);i!=0;i--)
   {
    switch(prg_path[i])
    {
     case '\\':
     case '/':
     case ':': prg_path[i+1]=0; goto avp_path_done;
    }
   }
  prg_path[0]=0;
avp_path_done:

  srcfd=open(AUDIT_SRC_PATH,O_RDONLY); 
  if(srcfd<0)
   {
    sprintf(buf,"Cannot open audit file(%s)\n",AUDIT_SRC_PATH);
    die(buf);
   }
   
  SetDefaultOptions(); 
  if(LoadProfile(AUDIT_CONF_PATH)==0)
   {
    LOG((1,"Cannot load profile. Profile is not loaded\n"));
    //die(buf);
   }
   
  char *tmp_c=CmdOpts;
  while(*tmp_c)
    {
     switch(toupper(*++tmp_c))
      {
       case 'E': 
	if((*(tmp_c+1))=='-') { tmp_c++; prf_data.UseTime=1;}
        else prf_data.UseTime=0;
        break; 
       case 'R': 
	if((*(tmp_c+1))=='-') { tmp_c++; prf_data.Directories=1;}
        else prf_data.Directories=0; 
        break; 
       case 'O': 
	if((*(tmp_c+1))=='-') { tmp_c++; prf_data.ShowOK=0;}
        else prf_data.ShowOK=1;
	break;         
       case 'S': 
	//if((*(tmp_c+1))=='-') { tmp_c++; prf_data.Sound=1;}
        //else prf_data.Sound=0;
	break;          
       case '-': 
         if(toupper(*(tmp_c+1))=='H') { tmp_c++; f_help=1;
 }
         break;
       case 'F':                                   //yes
         char defprofile[0x200];
         if(*++tmp_c=='=')
          {
           if(strchr(++tmp_c,'-')) len=(unsigned char)(strchr(tmp_c,'-')-tmp_c);
            else len=(unsigned short)strlen(tmp_c);
           strncpy(defprofile,tmp_c,len);
           defprofile[len]=0;
           LoadProfile(defprofile);
          }
        break;
      case 'X':
        ++tmp_c;
        switch(toupper(*tmp_c))
         {
          case 'F':
             if(*++tmp_c=='=')
              {
               prf_data.ExcludeFiles=1;
               if(strchr(++tmp_c,'-')) len=(unsigned char)(strchr(tmp_c,'-')-tmp_c);
               else len=(unsigned short)strlen(tmp_c);
               if((len>0)&&(tmp_c[0]=='{'))
                {
                 strncpy(prf_data.ExcludeMask,&tmp_c[1],len-1);
                 prf_data.ExcludeMask[len-1]=0;
                }
               else
                {
                 strncpy(prf_data.ExcludeMask,tmp_c,len);
                 prf_data.ExcludeMask[len]=0;
                }
               if(prf_data.ExcludeMask[strlen(prf_data.ExcludeMask)-1]=='}')
                prf_data.ExcludeMask[strlen(prf_data.ExcludeMask)-1]=0;
              }
             break;
          case 'D':
            if(*++tmp_c=='=')
             {
              prf_data.ExcludeFiles=1;
              if(strchr(++tmp_c,'-')) len=(unsigned char)(strchr(tmp_c,'-')-tmp_c);
              else len=(unsigned short)strlen(tmp_c);
              char tmpExclDir[0x200],*z0;
              strncpy(tmpExclDir,tmp_c,len);
              tmpExclDir[len]=0;
              z0=strtok(tmpExclDir,"; ,{}");
              while(z0)
               {
                if(z0[strlen(z0)-1]=='\\') z0[strlen(z0)-1]=0;
                //AddRecord(z0,2);
                z0=strtok(0,"; ,{}");
               }
             }
            break;
         }
        break;
       case '?':
         f_help=1;
         break;
       case 'W':                                     //yes
         prf_data.CommonLog=1;
         while((*tmp_c)&&(*tmp_c!='/'))
          switch(toupper(*++tmp_c))
           {
            case 'T': prf_data.Append=0; break;
            case 'A': prf_data.Append=1; break;
            //case '-': prf_data.ExtReport=0; break;
            //case '+': prf_data.ExtReport=1; break;
            case '=':
              if(strchr(++tmp_c,'-')) len=(unsigned char)(strchr(tmp_c,'-')-tmp_c);
              else len=(unsigned char)strlen(tmp_c);
              strncpy(prf_data.CommonLogFileName,tmp_c,len);
              prf_data.CommonLogFileName[len]=0;
              break;
           }
         break;
//     case 'Z': 
//	if((*(tmp_c+1))=='-') { tmp_c++; cmdline.abort_disable=0;}
//        else cmdline.abort_disable=1;
//	break;
     case 'L':     
         switch(toupper(*(tmp_c+1)))
          {
           case 'P': tmp_c++; prf_data.Symlinks=0; break;
 //no symlinks are followed
           case 'L': tmp_c++; prf_data.Symlinks=2; break;
 //all symlinks are followed
           case 'H':   //symlinks on the command line are followed
           default : tmp_c++; prf_data.Symlinks=1; break;

          }
         break;
     case 'C':
	if((*(tmp_c+1))=='-') { tmp_c++; prf_data.CrossFs=1;}
        else prf_data.CrossFs=0;
	break;   
      } // switch(toupper(*++tmp_c))
     while(*tmp_c&&*tmp_c!='-') tmp_c++;
    } // while(*tmp_c)

  prf_data.CommonLogfd=reportInit(&prf_data.CommonLog,prf_data.CommonLogFileName);

  if((DaemonPid=CheckPresent())<0) // not exists
   {
    return -1;
   }
   
  struct Message currMess;

  if(f_help==1)
   {
    usage();
    exit(0);
   }    
#ifdef __DEF_FORK__
  if(!fork())
#endif
    while(1)
     {
      bzero(&buf, sizeof(buf));
      count=getline(srcfd,buf,sizeof(buf));
      if(count)
       {
        //printf("Current message:{%s}\n",buf);
        getStructByLine(buf,&currMess);
        LOG((1,"Current message: flag: %c pname: %s pid: %d uid: %d {%s}\n",
	  currMess.flag,currMess.pname,currMess.pid,currMess.uid,currMess.buf));
        if(cmpFileMask(currMess.buf)==FALSE) 
         logline(currMess);
       }
     }
  reportDone(prf_data.CommonLog,prf_data.CommonLogfd);

  exit(0);
 }

